/* 
 * Copyright 1992, 1993, 1994 The Board of Trustees of the Leland Stanford
 * Junior University. Copyright 1994, 1995 Dave Nichols and Steve Cole.
 * Permission is hereby given to use, copy, modify, and
 * distribute this software, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * STANFORD AND THE AUTHORS MAKE NO REPRESENTATIONS OR WARRANTIES OF ANY 
 * KIND CONCERNING  THIS SOFTWARE.
 */

/*
 *
 *  source file:   ./xtpanel/actions.c
 *
 * Steve Cole, Dave Nichols (SEP), August 28 1992
 *      Inserted this sample edit history entry.
 *      Please log any further modifications made to this file:
 * Steve Cole, Dave Nichols (SEP), November 20 1992 
 *      new actions: ASSIGN, SET.
 *      objects can have multiple actions.
 * 	    backquoted strings in actions get executed at action time.
 *  Dave Nichols (SEP), January 19 1993 
 *      Use expand_string to evaluate actions at runtime, this avoids
 *      duplicate code and there is much less parsing to do.
 *  Dave Nichols (SEP), April 25 1993
 *      Introduced conditional actions "test" objects.
 *      QUIT now can have an action format which is passed through atoi to 
 *      generate a return code.
 *  Dave Nichols (SEP), May 7 1993
 *      Fixed all_string_actions to handle conditionals.
 *  Steve Cole (SEP), June 7 1993
 *      Added new actions CREATE and DESTROY.
 *  Dave Nichols (SEP), June 15 1993
 *	Added new routine one_word() and changed to new object style.
 *  Dave Nichols (SEP), July 11 1993
 *	Added new action EVAL
 *	Added check of return codes from action routines.
 *  Dave Nichols (SEP), April 11 1994
 *	Added USER functions.
 *  Dave Nichols (SEP), April 12 1994
 *	Added free_action routine.
 *  Dave Nichols (Geco-Prakla), January 30 1995
 *	Fixed memory leak in EVAL actions.
 *  Dave Nichols (Geco-Prakla), Aug 22 1995
 *	Defer deletion of actions until they have all been executed.
 *      This is to prevent a DESTROY action from deleting the active action list.
 *  Dave Nichols (Geco-Prakla), Jan 10 1996
 *	Don't link together actions in defreed deletion list they may be
 *	in the process of being executed!
 */

#include <stdio.h>
#include <string.h>

#include "tree.h"

#include "actions.h"
#include "object.h"
#include "eval_command.h"
#include "string_buf.h"
#include "expand_string.h"


/*
 * Function name: parse_actions
 * Description: parses all the actions for an object
 * Arguments: name (object->name) root (level in tree)
 * Returns: pointer to new structure of type action
 */
action* parse_actions( name, root )
     char *name;
     entry *root;
{
    action *topact;
    char *act_string;
    action *act;
    entry *ent;
    
    /* Initially no actions */
    topact = (action*) 0;
    
    /* loop over action strings, parse each */
    for (ent = get_next_act_or_test(root,(entry*)0);
	 ent != ((entry*) 0);
	 ent = get_next_act_or_test(root,ent)) {
	
	/* create a new action struct */
	if( topact == (action*) 0){
	    act = topact =  new_action( 0 );
	}else{
	    act = (action*) new_action(topact);
	}
	
	if( !strcmp( ent->tag, "test" )){
	    /* parse a test object from the current tree node */
	    parse_one_test( act, name, ent);
	    
	}else{
	    /* parse an action from the current tree node */
	    /* NOTE: don't use get_value to get the string because that
	     * will expand out the action string and we don't want this 
	     * to happen until action time */
	    if( ent != (entry*)0 &&  ent->value != (char*)0 ) {
		act_string = XtNewString( ent->value );
	    }else{
		act_string = XtNewString( "NONE" );
	    }
	    parse_one_action( act, name, act_string );
            XtFree( act_string );
	}
    }
    
    /* make an action of "NONE" if none were specified */
    if( topact == 0 ){
	act = topact =  new_action( 0 );
	parse_one_action( act, name, "NONE" );
    }
    
    /* return the head of the tree */
    return topact;
}
/*
 * Function name: parse_one_test
 * Description: parses a single test for an object
 * Arguments: act (action structure) name (object->name) 
 * root (tree node for test)
 * Returns: none
 */
void parse_one_test( act, name, root )
     action *act;
     char* name;
     entry* root;
{
    entry *true, *false, *cond;
    
    /* get the condition */
    /* don't use get_value because we don't want the string expanded */
    cond = get_next( root, "condition", (entry*)0 );
    if( cond != (entry*)0 && cond->value != (char*)0 ) {
	act->condition=XtNewString( cond->value );
    }else{
	fprintf( stderr,"A test object must have a condition.\n");
	fprintf(stderr,"The test object called \"%s\" does not.\n\n",name);
	exit(-1);
    }
    
    
    /* find the "true" entry */
    true = get_next( root, "true", (entry*)0 );
    if( true == 0 ){ 
	act->true_acts =0;
    }else{ 
	act->true_acts = parse_actions( name, true );
    }	  
    /* find the "false" entry */
    false = get_next( root, "false", (entry*)0 );
    if( false == 0 ){ 
	act->false_acts=0;
    }else{
	act->false_acts = parse_actions( name, false);
    }
}

/*
 * Function name: one_word
 * Description: extract one word from the line and return a pointer to the
 * position after the word. The word is returned in char** word.
 * A word is defined by separating space or tab.
 * (like strtok, but we don't want the line modified).
 */
static char* one_word( line, word )
char* line;
char** word;
{
char* startpos;
int len;

/* skip leading white space */
for(startpos=line; *startpos==' ' || *startpos == '\t' ; startpos++);  

/* find the end of the word */
len = strcspn( startpos, " \t" ); 

/* allocate and copy */
*word = (char*)XtMalloc( len + 1); 
strncpy( *word, startpos, len ); 
(*word)[len]='\0'; 

/* return pointer to next word or end of string*/
if( len == strlen( startpos ) ){
   return(startpos+len);
}else{
   return(startpos+len+1);
}

}

/*
 * Function name: parse_one_action
 * Description: parses a single action string for an object
 * Arguments: act (action structure) name (object->name) act_string (to parse)
 * Returns: none
 */
void parse_one_action( act, name, act_string )
     action *act;
     char* name;
     char* act_string;
{
    int len;
    
    /* strip leading blanks */
    len = strspn( act_string, " \t\n" );
    act_string += len;
    
    
    /* no action specified or "NONE" explicitly specified */
    if( !strncmp( act_string, "NONE" , 4 ) ){
  	act->type=NONE; return;
    }
    
    /* "QUIT" explicitly specified, the format string should be something
     * that atoi can convert to a number this will be our return code */
    if( !strncmp( act_string, "QUIT" , 4 ) ){
	act->type=QUIT; 
	act_string += 4;
	
    }else if( !strncmp( act_string, "PRINT" , 5 ) ){
	act->type=PRINT;
	act_string += 5;
	
    }else if(!strncmp( act_string, "STRING" , 6 ) ){
	act->type=STRING;
	act_string += 6;
	
    }else if(!strncmp( act_string, "ASSIGN" , 6 ) ){
	act->type=ASSIGN;
	act_string += 6;
	/* the next word is the object name to assign it to */
	act_string = one_word( act_string, & act->assignto );
	
    }else if(!strncmp( act_string, "SET" , 3 ) ){
	act->type=SET;
	act_string += 3;
	/* the next word is the object name to assign it to */ 
	act_string = one_word( act_string, & act->assignto );
	/* the next word is the tag to assign it to */
	act_string = one_word( act_string, & act->totag );
	
    }else if(!strncmp( act_string, "SYSTEM" , 6 ) ){
	act->type=SYSTEM;
	act_string += 6;

    }else if(!strncmp( act_string, "DESTROY" , 7 ) ){
	act->type=DESTROY;
	act_string += 7;
	act_string = one_word( act_string, & act->assignto );
	
    }else if(!strncmp( act_string, "CREATE" , 6 ) ){
	act->type=CREATE;
	act_string += 6;
	act_string = one_word( act_string, & act->assignto );
	
    }else if(!strncmp( act_string, "EVAL" , 4 ) ){
	act->type=EVAL;
	act_string += 4;
	
    }else if(!strncmp( act_string, "USER" , 4 ) ){
	char *fn_name;
	act->type=USER;
	act_string += 4;
        /* find the user function info */
        act_string = one_word( act_string, &fn_name );
        if( (act->user_func = find_function( fn_name )) == 0 ){
 	    fprintf(stderr,"Unknown user function name: \"%s\" \n", fn_name );
	    exit(-1);
        }
        /* type 2 functions take another object name as the second argument */
        if( act->user_func->type == 2 ){
	   act_string = one_word( act_string, & act->assignto );
        }
	
    }else {
	act->type=SYSTEM;
    }
    
    act->format = XtNewString( act_string );
    
}

/*
 * Function name: new_action
 * Description: creates a structure for a new action
 * Arguments: action* topact ( first item in action list for an object)
 * Returns: pointer to new structure of type action
 */
action* new_action(topact)
     action* topact;
{
    action *a, *b;
    
    /* make a new action */ 
    a = (action*) XtMalloc( sizeof( action ) );
    a->format = (char *) 0;
    a->assignto = (char *) 0;
    a->totag = (char *) 0;
    a->next= (action*) 0;
    a->false_acts=(action*)0;
    a->true_acts=(action*)0;
    a->condition=(char*)0;
    a->user_func=(user_fn*)0;
    
    if( topact == (action*)0 ) return a;
    
    /* find the tail of the list */
    for( b=topact; b->next != (action*)0; b = b->next );
    
    b->next = a;
    
    return a;
}

/* 
 * Static variables to control freeing of actions.
 * If "in_perform_actions" is true they will just be added to the
 * deferred deletion list (dead_actions).
 */
static int in_perform_actions=0;

struct dead_list {
	action* dead_one;
	struct dead_list* next;
};

struct dead_list* dead_actions=0;

/* Add the current action to the list of actions scheduled for later deletion */
static void deferred_delete( act )
action* act;
{
   struct dead_list* new_dead =
      (struct dead_list*)XtMalloc( sizeof(struct dead_list) );
   new_dead->dead_one = act;
   new_dead->next = dead_actions;
   dead_actions = new_dead;

   return;
}

static void process_deferred(){

struct dead_list* tmp;
while(dead_actions != 0 ){
    free_action( dead_actions->dead_one );
    tmp = dead_actions;
    dead_actions = dead_actions->next;
    XtFree( tmp );
}
   
}

/*
 * Function name: free_action
 * Description: frees memory associated with an action structure
 * Arguments: action* act ( first item in action list for an object)
 */
void free_action(act)
     action* act;
{
    if( act == 0 ) return;

    if( in_perform_actions !=0 ){
	deferred_delete( act );
	return;
    }

    if( act->assignto != 0) XtFree( act->assignto );  act->assignto=0;
    if( act->format != 0) XtFree( act->format ); act->format=0;
    if( act->totag != 0) XtFree( act->totag ); act->totag=0;
    if( act->condition != 0) XtFree( act->condition ); act->condition=0;

    if( act->false_acts != 0) free_action(  act->false_acts ); act->false_acts = 0;
    if( act->true_acts != 0) free_action(  act->true_acts ); act->true_acts=0;
    if( act->next != 0) free_action(  act->next ); act->next=0;

    XtFree( (char*)act );
    
}

/*
 * Function name: perform_actions
 * Description: performs all the actions for an object
 * Arguments: name (object->name) topact (head of action list) 
 *            all (>0 do all actions; =0 string actions only)
 * Returns: int
 */
extern void flush_display();

int perform_actions( name, topact, all )
     char *name;
     action *topact;
     int all;
{
    action *act;
    int retcode=0;
    char* savename;

    /* save the name in case an action delete the string */
    savename = XtNewString( name );


    /* flag to indicate level in action processing */
    in_perform_actions++;
    
    for (act=topact; act != (action*) 0; act=act->next ) {

	if( all && act->condition != 0 ){
	    /* this is a test, evaluate the condition */
	    if( eval_condition( savename, act->condition ) == 0 ){
		if( act->true_acts != 0 ) 
		  retcode += perform_actions( savename, act->true_acts, all );
	    }else{
		if( act->false_acts != 0 ) 
		  retcode += perform_actions( savename, act->false_acts, all );
	    }
	    
	}else{
	    /* this is an unconditional action, so do it */
	    if (all || act->type == STRING || act->type == NONE)
	      retcode += perform_one_action( savename, act );
	    /*flush_display();*/
	}
	
    }
    
    in_perform_actions--;

    /* We have completed all the actions in the list (and any derived lists)
     * It is now safe to clean up all the action lists that were logically
     * deleted as a result of those actions */
    if(  in_perform_actions==0  ){
	 process_deferred();	 
    }

    XtFree(savename);
    return retcode;
}

int eval_condition( name, cond )
     char*name;
     char* cond;
{
    char* condline;
    
    int ret;
    /* expand out the condition string */
    condline = expand_string( name, cond );
    /* evaluate the command and return the status */
    ret = system( condline );
    XtFree( condline );
    return ret;
}

/*
 * Function name: perform_one_action
 * Description: performs a single action for an object
 * Arguments: name (object->name) act (action) 
 * Returns: int
 */
int perform_one_action( name, act )
     char *name;
     action *act;
{
    char *action_line;
    char *action_target;
    char *action_tag;
    action *eval_act;
    extern void process_tree();
    
#ifdef ACT_DEBUG
    print_action( name, act );
#endif

    switch( act->type ){
	case( NONE ):
	  return 0;
	  break;
	
	case( QUIT ):
	  action_line = expand_string( name, act->format );
	  quit_xtpanel( atoi(action_line) );
	  break;
	
	case( PRINT ):
	  action_line = expand_string( name, act->format );
	  printf( "%s",action_line );
	  fflush(stdout);
	  XtFree( action_line );
	  return 1;
	  break;
	
	case( STRING ):
	  action_line = expand_string( name, act->format );
	  if( !set_string( name, action_line ))
 	    fprintf(stderr,"Error in STRING action for object \"%s\"\n",name);
	  XtFree( action_line );
	  return 0;
	  break;
	
	case( ASSIGN ):
	  action_target = expand_string( name, act->assignto );
	  action_line = expand_string( name, act->format );
	  if( !update_object( action_target, action_line ) )
 	    fprintf(stderr,"Error in ASSIGN action for object \"%s\"\n",name);
	  XtFree( action_target );
	  XtFree( action_line );
	  return 0;
	  break;
	
	case( SET ):
	  action_target = expand_string( name, act->assignto );
	  action_tag = expand_string( name, act->totag );
	  action_line = expand_string( name, act->format );
	  if( !update_tag( action_target, action_tag, action_line ) )
 	    fprintf(stderr,"Error in SET action for object \"%s\"\n",name);
	  XtFree( action_target );
	  XtFree( action_tag );
	  XtFree( action_line );
	  return 0;
	  break;
	
	case( SYSTEM ):
	  action_line = expand_string( name, act->format );
	  system( action_line );
	  XtFree( action_line );
	  return 1;
	  break;

	case( DESTROY ):
	  action_line = expand_string( name, act->assignto );
	  if( !destroy_object( action_line ) )
 	    fprintf(stderr,"Error in DESTROY action for object \"%s\"\n",name);
	  XtFree( action_line );
	  return 0;
	  break;
	
	case( CREATE ):
	  action_target = expand_string( name, act->assignto );
	  action_line = expand_string( name, act->format );
          if( !create_object( action_target,action_line ) )
 	    fprintf(stderr,"Error in CREATE action for object \"%s\"\n",name);
	  XtFree( action_target );
	  XtFree( action_line );
	  return 0;
	  break;
	
	case( EVAL ):
	  action_line = expand_string( name, act->format );
	  eval_act = new_action( 0 );
          parse_one_action( eval_act, name, action_line );
	  XtFree(action_line);
	  perform_one_action( name, eval_act );
	  free_action( eval_act );
	  return 0;
	  break;
	
	case( USER ):
	  action_line = expand_string( name, act->format );
	  if( act->user_func->type == 1 ){
	      act->user_func->func1( name, action_line, 
				act->user_func->user_data);  
          }else{
	      action_target = expand_string( name, act->assignto );
	      act->user_func->func2( name, action_target,action_line, 
				act->user_func->user_data);  
	      XtFree( action_target );
          }
	  XtFree(action_line);
	  return 0;
	  break;
    }
    
    return 0;
}

/*
 * Function name: all_string_actions
 * Description: determines whether all actions for an object are string type
 * Arguments: topact (head of action list )
 * Returns: int (=0 if not all actions are strings, =1 otherwise)
 */
int all_string_actions( topact )
     action *topact;
{
    action *act;
    int all_strings = 1;
    for (act=topact; act != (action*) 0; act=act->next) {
	if( act->condition != 0 ){
            if( act->true_acts != 0 ) 
		all_strings *= all_string_actions( act->true_acts );
            if( act->false_acts != 0 ) 
		all_strings *= all_string_actions( act->false_acts );
	}else{
	    if (act->type == PRINT || act->type == SYSTEM || 
	        act->type == ASSIGN || act->type == SET) all_strings = 0;
        }
    }
    return all_strings;
}



void print_action( name, act )
     char *name;
     action* act;
{
    char* expstr;
    
    if( name != 0 ) printf(" name: %s ,",name);
    switch( act->type ){
	case( NONE ):	
	  printf(" action type NONE\n ");
	return;
	break;
	case( QUIT ):
	  printf(" action type QUIT, ");
	return;
	break;
	
	case( PRINT ):
	  printf(" action type PRINT ");
	break;
	
	case( SYSTEM ):
	  printf(" action type SYSTEM, ");
	break;
	
	case( STRING ):
	  printf(" action type STRING, ");
	break;
	
	case( ASSIGN ):
	  printf(" action type ASSIGN, ");
	break;
	
	case( SET ):
	  printf(" action type SET, ");
	break;

	case( DESTROY ):
	  printf(" action type DESTROY, ");
	break;

	case( CREATE ):
	  printf(" action type CREATE, ");
	break;

	case( EVAL ):
	  printf(" action type EVAL, ");
	break;
    }

    if( act->assignto != 0) printf(" object of action: %s, ",act->assignto);
    if( act->totag != 0) printf(" attribute of action: %s ",act->totag);
    printf("\n");

    if( act->condition != 0) printf(" test condition: %s ", act->condition);
    if( act->false_acts != 0){printf(" false actions\n");print_action(name,act->false_acts);}
    if( act->true_acts != 0){printf(" true actions\n");print_action(name,act->true_acts);}
    
    if( act->format != 0 && strlen( act->format) != 0 ){
     	printf(" action format = ->%s<- \n",act->format);

        expstr = expand_string( name, act->format);    
        if( strcmp( act->format, expstr )!=0 ) {
		printf(" expanded format = ->%s<- \n", expstr );
        }
        XtFree( expstr );
    }
    printf("\n");

    
}
