/*
 * Copyright 1992, 1993, 1994 The Board of Trustees of the Leland Stanford
 * Junior University. Copyright 1994, 1995 Dave Nichols and Steve Cole.
 * Permission is hereby given to use, copy, modify, and
 * distribute this software, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * STANFORD AND THE AUTHORS MAKE NO REPRESENTATIONS OR WARRANTIES OF ANY 
 * KIND CONCERNING  THIS SOFTWARE.
 */


/*
 *
 *  source file:   ./xtpanel/graph.c
 *
 * Steve Cole, Dave Nichols (SEP), November 20 1992
 *      Inserted this sample edit history entry.
 *      Please log any further modifications made to this file:
 * Steve Cole (SEP), May 5 1993
 *	Rewrote to use form instead of box.
 *	Took out label and other annotation; better if users specify.
 * Steve Cole (SEP), May 13 1993
 *      Changed so that pixmap from objects behind graph will show
 *      through if graph has no pixmap of its own.
 * Steve Cole (SEP), June 3 1993
 *      Added graph_sync, allowing min and max values to be settable.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Scrollbar.h>

#include <stdio.h>
#include <assert.h>


#include "object.h" 
#include "tree.h"
#include "builders.h"
#include "string_buf.h"

#include <string.h>

typedef struct graphinfo {
    float minval;
    float maxval;
    char* format;
    int nsamp;
    Widget *widgets;
} _ginfo;

/* translation table used for graph widget */
char graph_trans[] =
  "<Leave>:         update_graph()";

extern void graph_scroll_callback();
extern void graph_update();
extern void graph_sync();
extern void parse_valuestring();

void build_graph(root,parent)
     entry *root;
     Objdef* parent;
{
    Objdef *object;
    char* label;
    struct graphinfo *graph_info;
    Arg args[20], orientarg[1];
    int narg, nargsv;
    Widget form, scrollbar, prev, parentw;
    float top;
    float *values;
    char *valstring,*tmp;
    int isamp;
    int width,height;
    char *orient;
    XtOrientation xtorient;
    char defname[12];
    static int numgraph=1;
    
    /* create new object */
    object = new_object(parent);
    parentw = parent->widgetname;

    /* construct default graph name */
    sprintf(defname,"graph%d",numgraph++);
    
    /* find name, action, nsamp in tree */
    object->name = get_value(root,"name",defname);
    label = get_value(root,"label",object->name);
    object->action = parse_actions(object->name,root);
    graph_info = (struct graphinfo*) XtMalloc( sizeof( struct graphinfo ) );
    object->info = graph_info;
    graph_info->minval = ((float) atof(tmp=get_value(root,"min","0")));XtFree(tmp);
    graph_info->maxval = ((float) atof(tmp=get_value(root,"max","1")));XtFree(tmp);
    graph_info->nsamp = ((int) atoi(tmp=get_value(root,"nsamp","10")));XtFree(tmp);

    /* form widget holds the scrollbars*/
    narg = 0;
    /* backgroundPixmap comes from parent if not specified */
    XtSetArg(args[narg], XtNbackgroundPixmap, ParentRelative); narg++;
    common_tags(parentw,root,args,&narg,
               SET_POSITION | SET_COLOR | SET_PIXMAP );
    /* no space in between scrollbars */
    XtSetArg(args[narg], XtNdefaultDistance, 0); narg++;

    form = XtCreateWidget("graphbox",formWidgetClass,parentw,args,narg);
    
    /* use translations to update value whenever mouse leaves the form */
    XtAugmentTranslations(form,XtParseTranslationTable(graph_trans));

    /* allocate space to hold slider values and widget names */
    values = (float*) XtMalloc ( graph_info->nsamp * sizeof(float));
    graph_info->widgets = (Widget*) 
              XtMalloc ( graph_info->nsamp * sizeof(Widget));

    /* get the string containing default values */
    valstring = get_value(root,"value","");

    /* decode */
    parse_valuestring(valstring,values,graph_info->nsamp," \n\t");

    graph_info->format = get_value(root,"format","%f");
    height = (int) atoi(tmp=get_value(root,"height","5"));XtFree(tmp);
    width = (int) atoi(tmp=get_value(root,"width","100"));XtFree(tmp);
    orient = get_value(root,"orientation","horizontal");

    narg = 0;
    /* common parameters */
    common_tags(form,root,args,&narg,SET_COLOR | SET_ORIENT);

    /* height, width */
    /* done here instead of in common_tags so we could have defaults */
    XtSetArg(args[narg], XtNlength, width); narg++;
    XtSetArg(args[narg], XtNthickness, height); narg++;
    /* turn off border */
    XtSetArg(args[narg], XtNborderWidth, 0); narg++;
    /* solid color thumb if foreground color requested */
    if (is_specified(root,"foreground")) {
       XtSetArg(args[narg], XtNthumb, None); narg++;
       }
    /* let pixmap of the form show through the scrollbars */
    XtSetArg(args[narg], XtNbackgroundPixmap, ParentRelative); narg++;

    nargsv = narg;

    /* loop over sliders */
    for (isamp=0; isamp<graph_info->nsamp; isamp++) {
  
    narg = nargsv;

    /* determine the correct starting point for the sliders */
    top = values[isamp]/graph_info->maxval;
    if (sizeof(float) > sizeof(XtArgVal))
      {
	  XtSetArg(args[narg], XtNtopOfThumb, top); narg++;
      }
    else
      {
	  XtArgVal * l_top = (XtArgVal *) &top;
	  XtSetArg(args[narg], XtNtopOfThumb, *l_top); narg++;
      }

    /* arrangement in form is relative to previous scrollbar
     * and depends upon orientation */
    if (isamp != 0) {
      XtSetArg(orientarg[0], XtNorientation, &xtorient); 
      XtGetValues( prev, orientarg, ONE );
      if (xtorient == XtorientHorizontal) {
    	XtSetArg(args[narg], XtNfromVert, prev); narg++;
      } else {
    	XtSetArg(args[narg], XtNfromHoriz, prev); narg++;
      }
    }

    assert ( narg < 20 );
    scrollbar = XtCreateManagedWidget(object->name,scrollbarWidgetClass,
                   form,args,narg);
    
    /* scroll callback is for incremental scrolling with left and 
       right buttons */
    XtAddCallback( scrollbar, XtNscrollProc, graph_scroll_callback, 
		  (XtPointer) object );
    
    /* set to the correct starting point */
    (void) XawScrollbarSetThumb(scrollbar,
    (values[isamp]-graph_info->minval)/
            (graph_info->maxval-graph_info->minval),-1.);
    
    graph_info->widgets[isamp] = scrollbar;
    prev = scrollbar;
    }
    
    /* now manage the form */
    XtManageChild(form);

    object->value = XtNewString(valstring);
    object->widgetname = form;
    object->updater = graph_update;
    object->sync = graph_sync;
    
    /* now hide object if needed */
    narg = 0;
    common_tags(form,root,args,&narg,SET_AFTER);

}

/* 
 * graph_scroll_callback 
 * called to do incremental scrolling of one of the sliders
 */
void
  graph_scroll_callback(widget, client_data, pos_ptr)
Widget widget;
XtPointer client_data, pos_ptr;
{
    Arg arg[1];
    float top;
    int pos;
    pos = (int) pos_ptr;
    /* get the current position of the graph */
    XtSetArg( arg[0], XtNtopOfThumb, &top );
    XtGetValues( widget, arg, ONE );
    /* now compute new position - 5% change */
    top -= pos/abs(pos) * 0.05;
    if (top > 1.) top = 1.;
    if (top < 0.) top = 0.;
    /* update the graph */
    if (sizeof(float) > sizeof(XtArgVal))
      {
	  XtSetArg(arg[0], XtNtopOfThumb, top);
      }
    else
      {
	  XtArgVal * l_top = (XtArgVal *) &top;
	  XtSetArg(arg[0], XtNtopOfThumb, *l_top);
      }
    XtSetValues( widget, arg, ONE );
}

/* 
 * graph_update 
 * called when the value of a graph is changed by an ASSIGN action 
 */
void
  graph_update(object, value)
Objdef *object;
char *value;
{
    struct graphinfo *graph_info;
    Arg arg[1];
    char text[64];
    float top;
    float *values;
    int isamp;
    string_buf *buffer;

    graph_info = (struct graphinfo *) object->info;
    values = (float*) XtMalloc ( graph_info->nsamp * sizeof(float));

    /* decode */
    parse_valuestring(value,values,graph_info->nsamp," \n\t");

    strcpy(object->value,"");
    buffer = buf_start();

    /* loop over sliders */
    for (isamp=0; isamp<graph_info->nsamp; isamp++) {

      top = (values[isamp] - graph_info->minval)/
             (graph_info->maxval-graph_info->minval);
      if (top > 1.) top = 1.;
      if (top < 0.) top = 0.;
      /* compute new value */
      sprintf( text, graph_info->format, 
	    (float) graph_info->minval +
	    top*(graph_info->maxval - graph_info->minval));
      buf_cat( buffer,text,strlen(text));
      buf_cat( buffer," ",1);
      /* update the graph */
      if (sizeof(float) > sizeof(XtArgVal))
      {
	  XtSetArg(arg[0], XtNtopOfThumb, top);
      }
      else
      {
	  XtArgVal * l_top = (XtArgVal *) &top;
	  XtSetArg(arg[0], XtNtopOfThumb, *l_top);
      }
      XtSetValues( graph_info->widgets[isamp], arg, ONE );
    }
    if( object->value != 0 ) XtFree( object->value );
    object->value = buf_fetch(buffer);
    buf_free(buffer);
}

/* 
 * graph_update_callback 
 * called whenever cursor leaves a graph box
 */
void graph_update_callback( w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
    Objdef* object;
    struct graphinfo *graph_info;
    Arg arg[1];
    int isamp;
    float top;
    char text[64];
    string_buf *buffer;

    object = find_by_widget(w);
    graph_info = (struct graphinfo *) object->info;

    strcpy(object->value,"");
    buffer = buf_start();

    /* loop over widgets */
    for (isamp=0; isamp<graph_info->nsamp; isamp++) {

    /* get the current position of the graph */
    XtSetArg( arg[0], XtNtopOfThumb, &top );
    XtGetValues( graph_info->widgets[isamp], arg, ONE );

    /* turn it into text, add to value string */
    sprintf( text, graph_info->format, 
	    (float) graph_info->minval +
	    top*(graph_info->maxval - graph_info->minval));
    buf_cat( buffer,text,strlen(text));
    buf_cat( buffer," ",1);
    }
    if( object->value != 0 ) XtFree( object->value );
    object->value = buf_fetch(buffer);
    buf_free(buffer);

    perform_actions(object->name,object->action,1);
}

/* 
 * parse_valuestring
 * parses a string containing a graph value to get the
 * values of the individual samples.
 */
void parse_valuestring(valstring,values,nsamp,separator) 
    char *valstring;
    float *values;
    int nsamp;
    char *separator;
{
    char *item;
    int isamp,jsamp;

    for( item=strtok(valstring,separator),isamp=0;
         item != (char*)0 && isamp <nsamp ;
         item =strtok( (char*)0, separator ),isamp++){
		sscanf(item,"%f",&values[isamp]);
    }

    /* if valstring does not contain enough samples, reuse the last one */
    if (isamp < nsamp) {
	for (jsamp=isamp; jsamp<nsamp; jsamp++) {
		values[jsamp] = values[isamp-1];
	}
    }
}

/*
 * graph_sync
 * called when a graph has a resource changed with SET. 
 * Setting of min and max is
 * handled here, because we must take care to update the
 * position of the scrollbars. Other resources are handled
 * by a call to update_common_tag.
 */
void graph_sync(object,tag,value)
Objdef* object;
char* tag;
char* value;
{
    struct graphinfo *graph_info;

    graph_info = (struct graphinfo *) object->info;

    /* most tags are handled in the usual way */
    update_common_tag(object,tag,value);

    /* handle min and max here */
    if (!strcmp(tag,"min"))
      graph_info->minval = ((float) atof(value));
    if (!strcmp(tag,"max"))
      graph_info->maxval = ((float) atof(value));

    /* call graph update to position scrollbar given new limits */
    graph_update(object,object->value);

}
