/*  
 * Copyright 1992, 1993, 1994 The Board of Trustees of the Leland Stanford
 * Junior University. Copyright 1994, 1995 Dave Nichols and Steve Cole.
 * Permission is hereby given to use, copy, modify, and
 * distribute this software, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * STANFORD AND THE AUTHORS MAKE NO REPRESENTATIONS OR WARRANTIES OF ANY 
 * KIND CONCERNING  THIS SOFTWARE.
 */


/*
 *
 *  source file:   ./xtpanel/input.c
 *
 *  Dave Nichols (SEP), July 3 1993
 *      Inserted this sample edit history entry.
 *      Please log any further modifications made to this file:
 *	Wrote input.c.
 *  Steve Cole (SEP), October 7 1993
 *      Added pipe and output objects.
 *  Dave Nichols (SEP), May 12 1994
 *      Added proper pclose of input and output to kill children.
 *  Dave Nichols (Geco-Prakla), Jan 30 1995
 *      Fixed memory leak in input buffering and made a separate input buffer
 *      for each object.
 *      Fixed memory leak in create_input()
 *  Dave Nichols (Geco-Prakla), Mar 13 1996
 *      Handle the case where the actions turn on/off the running flag.
 *      We have to keep processing the saved input if any separators are
 *      present.
 *      This changes the meaning of the running variable slightly. 
 *      It now means keep processing input. 
 *      If the fd is -1 it means that the stream actually finished.
 *  Dave Nichols (Geco-Prakla), April 16 1996
 *      Added a destroy proc to close pipes and cleanup child processes. 
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <stdio.h>
#include <math.h>

#include <sys/ioctl.h>
#include <sys/types.h>

#include <X11/Xos.h>

#ifndef STDIN_FILENO
#define STDIN_FILENO    0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO   1
#endif

#include "object.h" 
#include "tree.h"
#include "builders.h"
#include "string_buf.h"

extern Widget toplevel;

typedef struct inputinfo {
  char *type;
  unsigned long fd;
  unsigned long ofd;
  FILE* inpipe;
  FILE* outpipe;
  XtInputId id;
  int pipe_pid;
  int running;
  char *separator;
  string_buf *input_buffer;
} _inpinfo;

extern int stdin_used;
extern XtAppContext app_con;

extern void input_callback();
extern void input_sync();
extern void input_destroy();
extern void input_update();
static void process_buffer();

/* get the file descriptor to use for input or output */
static FILE* get_file( stream, objname, type  )
char* stream;
char* objname;
char* type;
{
  FILE* infile;
  FILE* outfile;

  if( !strcmp( stream, "none" )  ){
    return 0 ;
  }

  if( !strcmp( type, "input" ) ){
    if( !strcmp( stream, "stdin" )  ){
      /* use stdin if available */
      if( stdin_used ){
        fprintf(stderr,
		"input object %s cannot use stdin, already in use\n",objname);
	exit(-1);
      }
      stdin_used = 1;
      infile = stdin;

    }else{
      /* start the command */
      if( ( infile = popen( stream, "r" )) == 0 ){
	fprintf(stderr, "unable to execute input command %s for object %s\n", 
		stream,objname);
	exit(-1);
      }
    }
    return ( infile );
  } else {
    if( !strcmp( stream, "stdout" )  ){
      outfile = stdout;

    }else{
      /* start the command */
      if( ( outfile = popen( stream, "w" )) == 0 ){
	fprintf(stderr, "unable to execute output command %s for object %s\n", 
		stream,objname);
	exit(-1);
      }
    }
    return ( outfile );
  }
}

/* for object type pipe we need two pipes for two-way communication */
static void get_pipes( stream, objname, inpipe, outpipe, child_pid  )
char* stream;
char* objname;
int *inpipe, *outpipe,*child_pid;
{
  FILE* infile, *outfile;
  int pipe1[2], pipe2[2];
  int child ;

  if( !strcmp( stream, "none" )  ){
    *inpipe = -1;
    *outpipe = -1;
    return;
  }

  if( !strcmp( stream, "stdin" )  ){
    /* use stdin if available */
    if( stdin_used ){
      fprintf(stderr,
	      "pipe object %s cannot use stdin, already in use\n",objname);
      exit(-1);
    }
    stdin_used = 1;
    infile = stdin;
    *inpipe = fileno(infile);
    *outpipe = fileno(stdout);
    return ;

  }else{
    /* create the pipes */
    if( pipe(pipe1) < 0 || pipe(pipe2) < 0) {
      fprintf(stderr, "unable to open pipes for object %s\n", 
	      objname);
      exit(-1);
    }
 
    /* fork off a child process that will run the desired job */
    if ((child = fork()) == -1){
      perror("fork");
      exit(-1);

    }else if (child) {                       /* this is the parent */
      infile = fdopen(pipe1[0],"r");
      outfile = fdopen(pipe2[1],"w");
      /* close the unused pipe ends */
      if (close(pipe1[1]) != 0)
	fprintf(stderr,"parent: could not close pipe1[1]\n");
      if (close(pipe2[0]) != 0)
	fprintf(stderr,"parent: could not close pipe2[0]\n");
      *inpipe = fileno(infile);
      *outpipe = fileno(outfile);
      *child_pid = child;
      return; 

    } else {                                /* this is the child */
      /* reset stdin, stdout to the pipe file descriptors */
      if (close(STDIN_FILENO) != 0)
	fprintf(stderr,"could not close stdin\n");
      if (dup(pipe2[0]) != STDIN_FILENO)
	fprintf(stderr,"could not duplicate file descriptor\n");
      /* reset stdout */
      if (close(STDOUT_FILENO) != 0)
	fprintf(stderr,"child: could not close stdout\n");
      if (dup(pipe1[1]) != STDOUT_FILENO)
	fprintf(stderr,"child: could not duplicate file descriptor\n");

      /* close all the pipes */
      if (close(pipe1[0]) != 0)
	fprintf(stderr,"child: could not close pipe1[0]\n");
      if (close(pipe1[1]) != 0)
	fprintf(stderr,"child: could not close pipe1[1]\n");
      if (close(pipe2[0]) != 0)
	fprintf(stderr,"child: could not close pipe2[0]\n");
      if (close(pipe2[1]) != 0)
	fprintf(stderr,"child: could not close pipe2[1]\n");

      /* start the command */
      exit(system(stream));
    }
  }
  return ;
}

/* build an input, output, or pipe object */
void build_input(root,parent,type)
entry *root;
Objdef *parent;
char *type;
{
  Objdef *object;
  struct inputinfo *input_info;
  char defname[12];
  static int numinput=1;
  static int numoutput=1;
  static int numpipe=1;
  char *stream, *runflg;
  int inpipe, outpipe;
    
  /* create new object */
  object = new_object(parent);

  input_info = (struct inputinfo*) XtMalloc( sizeof(struct inputinfo) );
  input_info->inpipe =0;
  input_info->outpipe =0;
  input_info->fd =-1;
  input_info->ofd =-1;
  input_info->pipe_pid =0;
  input_info->input_buffer =0;

  /* construct default input name */
  if (!strcmp(type,"input"))
    sprintf(defname,"input%d",numinput++);
  else if (!strcmp(type,"output"))
    sprintf(defname,"output%d",numoutput++);
  else 
    sprintf(defname,"pipe%d",numpipe++);
    

  /* find label, name, action in tree */
  object->name = get_value(root,"name",defname);
  input_info->type = XtNewString(type);
  input_info->separator = get_value(root,"separator","\n");
  object->action = parse_actions(object->name,root);
  object->info = input_info;
  object->sync = input_sync;
  object->destroy = input_destroy;
  if (!strcmp(type,"output") || !strcmp(type,"pipe")) 
    object->updater = input_update;

  /* get the stream attribute */
  stream =  get_value(root,"stream","none");

  /* allocate file descriptors */
  if (!strcmp(type,"input")) {
    input_info->inpipe = get_file( stream, object->name, type );
    if( input_info->inpipe != 0 ){
      input_info->fd = fileno(input_info->inpipe);	
    }else{
      input_info->fd = -1;
    }
    input_info->ofd = -1;
  } else if (!strcmp(type,"output")) {
    input_info->fd = -1;
    input_info->outpipe = get_file( stream, object->name, type );
    if( input_info->outpipe != 0 ){
      input_info->ofd = fileno( input_info->outpipe );
    }else{
      input_info->ofd = -1;
    }
  } else {
    get_pipes( stream, object->name, &inpipe, &outpipe, &input_info->pipe_pid );
    input_info->fd = inpipe;
    input_info->ofd = outpipe;
  }

  XtFree( stream );

  input_info->id = 0;

  /* get the running attribute */
  if( (!strcmp(type,"input")  && (input_info->fd == -1) ) ||
      (!strcmp(type,"output") && (input_info->ofd == -1) ) ||
      (!strcmp(type,"pipe")   && ((input_info->fd == -1 ) ||
				  (input_info->ofd == -1 )))) {
    input_info->running = 0;
  }else{
    runflg = get_value(root,"running","True");
    input_info->running = get_boolean(parent->widgetname,runflg);
    XtFree( runflg );
  }

  /* add a callback for input available on this fd */
  if( input_info->running && (!strcmp(type,"input") ||
			      !strcmp(type,"pipe"))) {
    input_info->id = XtAppAddInput( app_con, input_info->fd, 
				    (XtPointer)XtInputReadMask, 
				    input_callback, (XtPointer)object );
  }
}

/* callback used for inputs and pipes */
void input_callback( client_data, source, id)
XtPointer client_data ;
int* source;
XtInputId* id;
{
  Objdef *object;
  struct inputinfo *input_info;

#define RBUFLEN 1024
  char readbuf[RBUFLEN];
  int nread,len;
    
  object = (Objdef *) client_data;
  input_info = (struct inputinfo*)object->info;

  if( input_info->id != *id ) {
    fprintf(stderr,"ID mismatch in input_callback\n");
    exit(-1);
  }

  if( input_info->input_buffer == 0 ) input_info->input_buffer = buf_start();
  /* 
   * We read to fill up readbuf (or however much is available )
   */

  nread = read( input_info->fd, readbuf, RBUFLEN-1 );

  if( nread < 0 ){
    perror("input_callback()");
    fprintf(stderr,"I/O error in input for object %s\n",object->name);
    exit(-1);
  }

  if( nread == 0 ){ /* EOF */
    /* no more input will be read */
    XtRemoveInput( input_info->id );
    input_info->id = 0;
    if( input_info->inpipe != 0 ){
      pclose(input_info->inpipe);
    }else{
      close(input_info->fd);
    }
    input_info->inpipe = 0;
    input_info->fd = -1;

    /* put a separator character in the readbuf */
    readbuf[0] = (input_info->separator)[0];
    nread = 1;
    /* then fall through so that anything remaining gets processed */
  }
    
  /* add the new input to the buffer */
  buf_cat( input_info->input_buffer, readbuf, nread );

  /* if there was a separator in the input we need to process the buffer*/
  if( strpbrk( readbuf, input_info->separator  ) != 0 ){
    process_buffer( object, input_info );
  }

}

static void process_buffer( object, input_info )
Objdef *object;
struct inputinfo *input_info;
{
  char *startpos, *pos, *endpos, *next;
  int len;

  if( input_info->input_buffer == 0 ) return;
  if( input_info->input_buffer->len == 0 ) return;
	
  /* gets string and resets buffer */
  pos = startpos = buf_fetch(input_info->input_buffer); 
  endpos = pos + strlen(pos);

  /* find the next separator */
  next = strpbrk( pos, input_info->separator );

  while( next != 0 ){
	   
    if( (len = next - pos ) != 0 ){ 

      /* copy the string to the object value and call the action */
      if( object->value != 0 ) XtFree( object->value );
      object->value = (char*)XtMalloc( len +1 );
      strncpy( object->value, pos, len );
      object->value[len] = '\0';
      /* Do the actions[s] */
      perform_actions(object->name,object->action,1);

    }

    /* we need to check if the action turned off this object */
    if( !input_info->running ) { 
      /* just update the pointer and stop processing */
      pos = next+1;
      break;
    }else{
      /* find the next separator */
      pos = next+1;
      next = strpbrk( pos, input_info->separator );
    }
  }

  /* anything left is put back at the start of the buffer */
  if( pos != endpos ){
    buf_cat( input_info->input_buffer, pos, strlen(pos) );
  }
  XtFree(startpos);

}

/*
 * this routine is called when a input's state (running
 * or not) or stream is changed by the SET action.
 */
void input_sync(object,tag,value)
Objdef* object;
char* tag;
char* value;
{
  struct inputinfo *input_info;
  int inpipe, outpipe;

  input_info = (struct inputinfo *) object->info;

  /* remove old input callback so we can start over */
  if( input_info->running && (!strcmp(input_info->type,"input") ||
			      !strcmp(input_info->type,"pipe"))) {
    if( input_info->id != 0 ) XtRemoveInput(input_info->id);
    input_info->id = 0;
  }

  if (!strcmp(tag,"running")) {
    input_info->running = get_boolean(object->parent->widgetname,value);
  }

  if (!strcmp(tag,"stream")) {
    input_info->id = 0;
    /* close up the old fd */
    if( input_info->inpipe != 0 ){
      pclose( input_info->inpipe );
      input_info->inpipe = 0;
    }else{
      if( input_info->fd != -1 ) close(input_info->fd);
    }
    input_info->fd = -1;
    if( input_info->outpipe != 0 ){
      pclose( input_info->outpipe );
      input_info->outpipe = 0;
    }else{
      if( input_info->ofd != -1 ) close(input_info->ofd);
    }
    input_info->ofd = -1;
    /* open a new stream */
    if (!strcmp(input_info->type,"input")) {
      input_info->inpipe = get_file( value, object->name, input_info->type );
      if( input_info->inpipe != 0 ){
	input_info->fd = fileno(input_info->inpipe);	
      }
    } else if (!strcmp(input_info->type,"output")) {
      input_info->outpipe = get_file( value, object->name, input_info->type);
      if( input_info->outpipe != 0 ){
	input_info->ofd = fileno( input_info->outpipe );
      }
    } else {
      get_pipes( value, object->name, &inpipe, &outpipe );
      input_info->fd = inpipe;
      input_info->ofd = outpipe;
    }
  }

  if(  !strcmp(input_info->type,"input") ||
       !strcmp(input_info->type,"pipe" )     ) {

    /* process anything remaining in the input buffer */
    process_buffer( object, input_info );

    if(  input_info->running && input_info->fd != -1 ){
      /* wait for more input */
      input_info->id = XtAppAddInput( app_con, input_info->fd, 
				      (XtPointer)XtInputReadMask, input_callback, 
				      (XtPointer)object);
    }
  } 
}

void input_update(object,value)
Objdef* object;
char* value;
{
  struct inputinfo *input_info;
  int nwrite;

  input_info = (struct inputinfo *) object->info;

  /* only a pipe or output object can write to the process */
  if (!strcmp(input_info->type,"input")) return;

  nwrite = write( input_info->ofd, value, strlen(value) );

  if( nwrite < strlen(value) ){
    perror("input_update()");
    fprintf(stderr,"I/O error trying to update object %s\n",object->name);
    exit(-1);
  }
}

void input_destroy(object)
Objdef* object;
{
  struct inputinfo *input_info;

  input_info = (struct inputinfo *) object->info;
  
  if( input_info->outpipe != 0 ){
     pclose( input_info->outpipe );
     input_info->outpipe = 0;
  }

  if( input_info->inpipe != 0 ){
     pclose( input_info->inpipe );
     input_info->inpipe = 0;
  }

  if( input_info->input_buffer != 0 ) {
     buf_free(  input_info->input_buffer );
     input_info->input_buffer = 0 ;
  }

  if( input_info->pipe_pid != 0 ) {
     input_info->pipe_pid = 0;
  }
  XtFree( input_info->type );
  XtFree( input_info->separator );
  
}
