/*
 * Copyright 1992, 1993, 1994 The Board of Trustees of the Leland Stanford
 * Junior University. Copyright 1994, 1995 Dave Nichols and Steve Cole.
 * Permission is hereby given to use, copy, modify, and
 * distribute this software, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * STANFORD AND THE AUTHORS MAKE NO REPRESENTATIONS OR WARRANTIES OF ANY 
 * KIND CONCERNING  THIS SOFTWARE.
 */


/*
 *
 *  source file:   ./xtpanel/object.c
 *
 * Steve Cole, Dave Nichols (SEP), August 31 1992
 *      Inserted this sample edit history entry.
 *      Please log any further modifications made to this file:
 * Steve Cole, Dave Nichols (SEP), November 20 1992 -  version 2.00
 *      objects can have multiple actions.
 * Steve Cole (SEP), January 14 1993
 *      Added call to object->sync after update_tag, to make sure
 *	that object values are kept up to date.
 * Steve Cole (SEP), June 7 1993
 *      Added new routines create_object and destroy_object.
 * Steve Cole (SEP), October 8 1993
 *      Added object->destroy for object-specific destruction routines.
 *      Modified update_tag routine and added update_common_tag to
 *      handle resources of composite widgets in a more sensible way.
 * Dave Nichols (GecoPrakla), Jan 10 1996
 *      Make sure all relevant desroy procs in an object tree are called.
 */

#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Viewport.h>

#include "object.h"
#include "tree.h"
#include "builders.h"

static Objdef* topobj=0;

/*
 * Function name: new_object
 * Description: allocates a structure for a new object
 * Arguments: none
 * Returns: pointer to new structure of type objdef
 */
Objdef* new_object( parent )
Objdef* parent;
{
    Objdef *a, *b;
    
    a = (Objdef*) XtMalloc( sizeof( Objdef ) );
    a->name= (char *)0;
    a->action= (action *)0;
    a->value= (char *)0;
    a->widgetname= (Widget)0;
    a->childwidget= (Widget)0;
    a->info= (void *)0;
    a->updater= 0;
    a->destroy= 0;
    a->sync= 0;
    a->tab= (char *)0;
    a->tab_proc= 0;
    a->parent = parent;
    a->child1 = (Objdef*)0;
    a->next_sib = (Objdef*)0;
 
    if( parent == 0 ){
	topobj = a; 
	return a;
    }  

    if( parent->child1 == 0 ){
	parent->child1 = a;
    }else{ 
       /* find the tail of the child list */
       for( b=parent->child1; b->next_sib != (Objdef*)0; b = b->next_sib );
       /* add me to the list */  
       b->next_sib = a;
    }
    
    
    return a;
}

/*
 * Function name: next_object
 * Get the next object in the tree.
 *
 */
static Objdef* next_object( obj )
Objdef* obj;
{
Objdef* curr;
  
  /* if I have children go down */  
  if( obj->child1 != 0 ){
	/* my children */
	return(obj->child1);
  }

  curr = obj;

  do{
      if( curr->next_sib != 0 ){
	/* more siblings */
	return(curr->next_sib);
      }

      /* otherwise back up to my parent */
      curr = curr->parent;

  }while( curr != 0 );

  /* if we got here we have searched the whole tree */
  return( (Objdef*) 0 );
	
}

/*
 * Function name: free_object
 * Description: free storage associated with an object and it's children.
 * Arguments: object - object to free, sibs - free siblings as well?
 * Returns: none
 */
static void free_object(object, sibs)
     Objdef* object;
     int sibs;
{
   if( object == 0 ) return;
   if( object->child1 != 0 )  free_object(object->child1,1);
   if( sibs && object->next_sib != 0 )  free_object(object->next_sib,1);

   if( object->tab != 0 ) XtFree(object->tab);
   if( object->name != 0 ) XtFree(object->name);
   if( object->value != 0 ) XtFree(object->value);
   if( object->info != 0 )  XtFree(object->info);
   if( object->action != 0 ) free_action( object->action );
   XtFree( (char*)object );
}   

/*
 * Function name: remove_object
 * Description: remove an object's structure from object tree.
 * Arguments: object - object to remove
 * Returns: none
 */
void remove_object(object)
     Objdef* object;
{
     Objdef* obj, *parent;

    /* loop over objects to find this one */
    for ( obj=topobj; obj != (Objdef*) 0; obj = next_object( obj ))
      {
	  if ( obj == object ) {
		parent = object->parent;
		if( parent->child1 == object ){
		   parent->child1 = object->next_sib;
                }else{
		   /* find the previous sib */
    		   for(obj=parent->child1;
			obj->next_sib != object;
			obj=obj->next_sib);
		   obj->next_sib = object->next_sib;
		}
		free_object(object,0);
		break;
	  }
      }
}

/*
 * Function name: find_by_name
 * Description: loop over all objects, find the one with
 *		the right name.
 * Arguments: name - name to match
 * Returns: pointer to object structure for matching object
 */
Objdef* find_by_name(name)
    char* name;
{
    Objdef* obj;
    
    /* loop over objects */
    for ( obj=topobj; obj != (Objdef*) 0; obj = next_object( obj ))
      {
	  if (!strcmp(obj->name,name)) {
	      return obj;
	  }
      }
    
    fprintf(stderr,"find_by_name(): cannot find object \"%s\"\n",name);
    return 0;
}

/*
 * Function name: find_widget
 * Description: loop over all objects, find the one with
 *		the right widget name.
 * Arguments: w - widget id to match
 * Returns: pointer to object structure for matching object
 */
Objdef*
  find_by_widget(w)
Widget w;
{
    Objdef* obj;
    
    /* loop over objects */
    for ( obj=topobj; obj != (Objdef*) 0; obj = next_object( obj ))
      {
	  if (obj->widgetname == w) {
	      return obj;
	  }
      }
    
    fprintf(stderr,"find_by_widget cannot find widget %d\n", (int)w );
    exit(-1);  
}

/*
 * Function name: get_string
 * Description: find the value of an object, given its name
 * Arguments: name - object name
 * Returns: pointer to string containing objects value
 */
char* get_string(name)
char* name;
{
Objdef *obj;
char * val;
    if( ( obj = find_by_name(name) ) == 0 ) return 0;
    if( (val = obj->value) == (char*) 0 ){
      return(XtNewString(""));  
    }else{
      return(XtNewString(val));  
    }
}

/*
 * Function name: set_string
 * Description: set the value of an object
 * Arguments: name - object name
 *	      string - pointer to string containing new object value
 * Returns: int true/false = succeeded/did not succeed
 */
int set_string(name,string)
     char* name,*string;
{
    Objdef *obj;
    if( ( obj = find_by_name(name) ) == 0 ) {
	fprintf(stderr,"Error in STRING action on %s\n",name);
	return 0;
    }
    
    if( obj->value != (char*)0 ) XtFree(obj->value);
    obj->value = XtNewString(string);
    return 1;
}

/*
 * Function name: all_actions
 * Description: do the actions of all objects whose action type
 *		is STRING. This makes sure defaults are set.
 * Arguments: none
 * Returns: none
 */
void all_actions()
{
    Objdef* obj;
    
    if( topobj == (Objdef*)0 ) return;
    
    /* loop over objects, skip the top one */
    for ( obj=topobj; obj != (Objdef*) 0; obj = next_object( obj ))
      {
	  if (obj->action != (action *) 0) {
		 perform_actions(obj->name,obj->action,0);
	  }
      }
}

/*
 * Function name: update_object
 * Description: update the value of an object
 * Arguments: name - object name
 *	      string - pointer to string containing new object value
 * Returns: int true/false = succeeded/did not succeed
 */
int update_object(name,string)
     char* name,*string;
{
    Objdef *object;

    if( ( object = find_by_name(name) ) == 0 ) {
	fprintf(stderr,"Error attempting to ASSIGN to %s\n",name);
	return 0;
    }

    if( object->value != (char*)0 ) XtFree(object->value);
    object->value = XtNewString(string);

    if( object->updater != 0 ) object->updater(object,string);
    return 1;
}

/*
 * Function name: update_tag
 * Description: update the value of an object's tag
 * Arguments: name - object name
 *            tag  - object tag that is changed.
 *	      value - pointer to string containing new object value
 * Returns: int true/false = succeeded/did not succeed
 */
int update_tag(name,tag,value)
     char* name,*tag,*value;
{
    Objdef *object;

    if( ( object = find_by_name(name) ) == 0 ) {
	fprintf(stderr,"Error attempting to SET attribute of %s\n",name);
	return 0;
    }

    if( object->sync == 0 ) {
        /* the default sync action is just to update the
         * corresponding widget resources */
	update_common_tag(object,tag,value);
    }else{

        /*   
         * object->sync is for objects who may need to do
         * something different after a SET command (such as toggles, whose
         * state may have changed). Or objects may have private
	 * attributes (such as the min and max values of scrollbars
	 * and sliders).
         * Note that object sync routines can still call update_common_tag
         * to handle the attributes that are not object specific.
         */
        object->sync(object,tag,value);
    }
    return 1;
}

/*
 * Function name: update_common_tag
 * Description: updates tags common to most objects
 * Arguments: object - object to be modified
 *            tag  - object tag that is changed.
 *	      value - pointer to string containing new object value
 * Returns: int true/false = succeeded/did not succeed
 * called by update_tag or by object-specific sync routines.
 */
int update_common_tag(object,tag,value)
     Objdef *object;
     char *tag,*value;
{
	int type;

	type = tag_type(tag);
        if( object->widgetname != 0){

	  /* 
	   * if type=0, apply only to object->widgetname 
	   */
	  if (type == 0) {
		  update_widget(object->widgetname,tag,value);
	  /* 
	   * if type=1, apply to object->widgetname,
	   * and to all children if widgetname is a composite.
	   * If widgetname is not a composite and childwidget
	   * is defined (e.g., menubuttons) apply to it - and
	   * to its children.
	   */
	  } else if (type == 1) {
		  update_widget(object->widgetname,tag,value);
		  if (XtIsComposite(object->widgetname)) {
	  		update_children(object->widgetname,tag,value);
		  } else if (object->childwidget != (Widget) 0) {
	  		update_widget(object->childwidget,tag,value);
		  	if (XtIsComposite(object->childwidget)) 
	  		  update_children(object->childwidget,tag,value);
		  }
	  /* 
	   * if type=-1, then
	   *   if widgetname is not composite, apply to it
	   *   if widgetname is composite, then:
	   *	 if childwidget defined, apply to it
	   *	 else apply to widgetname's children.  
	   */
	  } else if (type == -1) {
	       if (!XtIsComposite(object->widgetname)) {
		  update_widget(object->widgetname,tag,value);
	       } else {
	          if (object->childwidget != (Widget) 0) 
			update_widget(object->childwidget,tag,value);
		  else
	      		update_children(object->widgetname,tag,value);
	       }
	  }
	}
    return 1;
}

/*
 * Function name: update_widget
 * Description: updates the tags of a widget
 * Arguments: w - widget
 *            tag  - object tag that is changed.
 *	      value - pointer to string containing new object value
 *	      dokids - flag to indicate whether to also update children
 * Returns: int true/false = succeeded/did not succeed
 */
int update_widget(w,tag,value)
     Widget w;
     char *tag,*value;
{
        Arg args[20];
        int narg;

	narg = 0;
        SetTag(w,args,&narg,tag,value);
        XtSetValues(w,args,narg);

    return 1;
}

/*
 * Function name: update_children
 * Description: updates the tags of the children of a composite widget
 * Arguments: w - parent widget
 *            tag  - object tag that is changed.
 *	      value - pointer to string containing new object value
 * Returns: int true/false = succeeded/did not succeed
 * called by update_tag or by object-specific sync routines.
 */
int update_children(w,tag,value)
     Widget w;
     char *tag,*value;
{
        Arg args[20];
        int narg;
        Cardinal numchild;
        WidgetList children;
        Cardinal i;

        if (!XtIsComposite(w)) return 1;

	/* find number of children */
	narg = 0;
     	XtSetArg( args[narg], XtNnumChildren, &numchild ); narg++;
        XtSetArg( args[narg], XtNchildren, &children ); narg++;
        XtGetValues( w, args, narg );
        /* loop over children */
        for (i=0; i<numchild; i++)
        {
	    narg = 0;
            SetTag(w,args,&narg,tag,value);
            XtSetValues(children[i],args,narg);
	    /* call recursively if needed */
	    if (XtIsComposite(children[i])) 
		update_children(children[i],tag,value);
	}
    return 1;
}

/*
 * Function name: destroy_object
 * Description: destroy an object
 * Arguments: string - object name to destroy
 * Returns: int = succeed/did not succeed
 */
int destroy_object(string)
     char* string;
{
    Objdef *object;

    if( ( object = find_by_name(string) ) == 0 ) {
	fprintf(stderr,"Error attempting to DESTROY %s\n",string);
	return 0;
    }


    /* descend the tree of objects to ensure all is destroyed */
    if( object->child1 != 0 )  destroy_object(object->child1->name);

    /* first destroy the associate data then remove the object from the
     * object tree (this also frees the memory for the object */

    if( object->destroy == 0 ) {

      /* destroy the widget */
      if (object->widgetname != (Widget) 0) { 
         XtDestroyWidget(object->widgetname);
      }

    }else{

        /*   
         * object->destroy is for objects such as timers where
	 * destroy the widget is not appropriate, or windows
         * where there are multiple widgets to destroy,
         * or objects whose private info structure points to 
         * other data that must be freed.
         */
        object->destroy(object);
    }

    /* destroy the object structure */
    remove_object(object);
 
    return 1;
}

/*
 * Function name: create_object
 * Description: create an object
 * Arguments: name - object name of new object's parent
 *            string - script to build new object
 *	      root - entry to tree containing new object info
 * Returns: int succeed/did not succeed
 */
int create_object(name,string)
     char* name,*string;
{
    Objdef *object, *newobj;
    entry *root;
    Widget  child;
    struct compositeinfo *composite_info;
    Arg arg[1];
    extern Cardinal find_insert_position();

    object = find_by_name(name);
    if (object == (Objdef*) 0) object = find_by_name("window1");
    child = object->widgetname;

    /* if object is composite, ok. otherwise use its parent. */
    if (!XtIsComposite(child)) {
	newobj = object->parent;
        composite_info = (struct compositeinfo *) newobj->info;
        composite_info->insert_position = child;
        XtSetArg(arg[0], XtNinsertPosition, find_insert_position);
        XtSetValues( newobj->widgetname, arg, ONE );
    } else {
	newobj = object;
        composite_info = (struct compositeinfo *) object->info;
        composite_info->insert_position = (Widget) 0;
    }

    /* create new tree */
    root= new_entry( "root", 4);
    /* scan string and fill tree */
    xtpanel_scan( string,root );
    /* print new tree - for debugging */
    /*    print_entry( stderr,root,0 );*/
    /* process the new tree */
    process_tree( root,newobj );

    delete_entry( root );

    return 1;
}
