/*
 *	The following AT&T software is hereby put into the public domain:
 *
 *	    sciplot.c		[ this file ]
 *	    help/sciplot	[ online help text file ]
 *
 * 	This does not constitute a precedent to be applied to any other AT&T 
 *	copyright software.
 *
 *	AT&T relinquishes all intellectual property rights to this code, both
 *	source and binary form and permission is granted for anyone to use,
 *	duplicate, modify and redistribute it.
 *
 *	AT&T provides absolutely NO WARRANTEE OF ANY KIND with respect to
 *	this software. The entire risk as to quality and performance of this
 *	software is with the user.
 *
 *	IN NO EVENT WILL AT&T BE LIABLE TO ANYONE FOR ANY DAMAGES ARISING
 *	OUT OF THE USE OF THIS SOFTWARE, INCLUDING, WITHOUT LIMITATION,
 *	DAMAGES RESULTING FROM LOST DATA OR LOST PROFITS, OR FOR ANY
 *	SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES.
 *
 * 	THIS PACKAGE INCLUDES:
 *
 *	    Source code necessary to integrate Robert McMullen's SciPlot
 *	    plotting widget into xtpanel, and an online help text file
 *	    outlining its usage.
 *
 *
 * Original copyright notice pertaining to the xtpanel package:
 * 
 *	Copyright 1992, 1993, 1994 The Board of Trustees of the
 *	Leland Stanford Junior University. Permission is hereby given
 *	to use, copy, modify, and distribute this software, subject to
 *	the following conditions: The above copyright notice and this
 *	permission notice shall be included in all copies or substantial
 *	portions of the Software. STANFORD MAKES NO REPRESENTATIONS OR
 *	WARRANTIES OF ANY KIND CONCERNING THIS SOFTWARE.
 *
 *
 * Copyright notices pertaining to the SciPlot widget source code
 * can be found in files SciPlot.c, SciPlot.h, and SciPlotP.h.
 * No SciPlot widget source code appears in this source file.
 *
 */

/*
 * sciplot.c 1.20
 *
 * SciPlot widget handler for xtpanel. The SciPlot widget itself is the
 * work of Robert W. McMullen, rwmcm@orion.ae.utexas.edu.
 * 
 * See help/sciplot for a brief description of the script-level interface.
 *
 * After climbing the widget-integration learning curve, I went back
 * and retro-documented in this source file what all has to be done
 * to incorporate a new (generic) widget; consequently many of the
 * comments are of a tutorial nature, and refer generically to a "new"
 * widget, rather than the SciPlot widget specifically. Hopefully this
 * file can serve as a template for future widget-adders.
 *
 *
 *	- G. D. Golden
 *	  AT&T Bell Laboratories
 *	  Crawford Hill Laboratory
 *	  Holmdel, NJ
 * 
 */

/*
 *	E D I T   H I S T O R Y 
 *	=======================
 *
 *
 * Glenn Golden, (AT&T Bell Labs), July 3, 1995
 *	Inserted this sample edit history entry.
 *	Please log any further modifications made to this file.
 *
 * Glenn Golden (AT&T Bell Labs), July 6, 1995
 *	Cleaned up defaulting for label, xlabel, ylabel. Null string
 *	for any of these means "turn it off".
 *
 * Glenn Golden (AT&T Bell Labs), July 7, 1995
 *	Added polar plot. Selection via plottype parameter, For now,
 *	either "scope" or "polar".
 *
 * Glenn Golden (AT&T Bell Labs), July 10, 1995
 *	Added real sciplot_sync() function, to allow dynamic parameter
 *	updates via SET. 
 *
 * Glenn Golden (AT&T Bell Labs), July 14, 1995
 *	Autoscale Y axis if neither min nor max explicitly specified.
 *	
 * Glenn Golden (AT&T Bell Labs), July 15, 1995
 *	Added static plotting plottypes "cartplot" and "polarplot".
 *	Renamed "polar" to "polarscope" to make it more obvious
 *	(well, less obscure, anyway) that it operates in a scope-like
 *	"append" mode rather than "replace the whole dataset" mode.
 *	For naming symmetry, original "scope" mode can now be referred
 *	to as either "scope" or "cartscope".
 *	
 * Glenn Golden (AT&T Bell Labs), Sept 27, 1995
 * 	Added AT&T legal statement and cleaned up comments.
 * 
 * Glenn Golden (Bell Labs, Lucent Technologies), Mar 1, 1996
 *	Update function now exploits SciPlotQuickUpdate() for
 *	cartscope and polarscope object types. This function is
 *	new in SciPlot-1.31, so this version and subsequent ones
 *	(1.16 and later) are no longer backward compatible with
 *	SciPlot-1.2. Tradeoff for improved realtime performance is
 *	no autoscaling in scope modes.
 *
 * Glenn Golden (Bell Labs, Lucent Technologies), Mar 2, 1996
 *	Fixed parsing error in sciplot_update(), POLARSCOPE mode.
 *	(ASSIGNed R,theta pairs separated by commas got botched.)
 *
 * Glenn Golden (Bell Labs, Lucent Technologies), May 10, 1996
 *	Added linestyle and marker attributes to sciplot. Requires
 *	SciPlot-1.34.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Form.h>
#include <stdio.h>
#include <math.h>

/*
 * xtpanel headers.
 */
#include "object.h" 
#include "tree.h"
#include "builders.h"

/*
 * Header for the widget(s) upon  which the new object is based.
 */
#include "SciPlot.h"			/* SciPlot widget header file. */

/*
 * Default values, data sizing, max's & min's.
 */ 
#define	OBJNAME		"sciplot"	/* Name of object. */

#define	YMIN_DEFAULT	0.0		/* Defaults for min... */
#define	YMAX_DEFAULT	1.0		/* ...and max. */

#define	NSAMP_MAX	1000		/* Max # point pairs (arbitrary). */
#define	NSAMP_DEFAULT	"10"		/* Default for nsamp. */

/*
 * Line styles and point marker types.
 */
#define LS_NONE		XtLINE_NONE
#define LS_SOLID	XtLINE_SOLID
#define LS_DOT		XtLINE_DOTTED
#define LS_WDOT		XtLINE_WIDEDOT 

#define	MKR_NONE	XtMARKER_NONE
#define	MKR_DOT		XtMARKER_DOT
#define	MKR_CIRC	XtMARKER_CIRCLE
#define	MKR_SQR		XtMARKER_SQUARE
#define	MKR_TRI		XtMARKER_UTRIANGLE
#define	MKR_FCIRC	XtMARKER_FCIRCLE
#define	MKR_FSQR	XtMARKER_FSQUARE
#define	MKR_FTRI	XtMARKER_FUTRIANGLE

static	char SCCSID[] = "@(#)sciplot.c	1.20 - AT&T Bell Laboratories";

/*
 * Enumerator for the possible values of the "plottype" tag:
 *
 * SCOPE	An oscilloscope-like display (rectangular coordinates).
 * 		X axis is like time; values assigned to the object are taken
 *		to be successive abcissas, with new values progressively
 *		added to right side of the plot, and the older values falling
 *		off the left side. Autoscaling is not performed in order to
 *		to keep the repaint snappy.
 *
 *		With both server and client running on the same Sparc LX,
 *		SCOPE mode updates reasonably smoothly up to about 50
 *		points/second.
 *
 *		For symmetry with the other plot types, the plottype tag value
 *		for a SCOPE can be specified as either "scope" or "cartscope".
 *
 * POLARSCOPE	Similar to SCOPE, except no implicit time axis. Values are
 *		taken to be (R, theta) pairs, with R in radians. The most
 *		recent nsamps points are drawn, connected. Note that
 *		SciPlot always seems to autoscale when in polar mode, so
 *		that's what you get - min and max don't have any effect.
 *
 *		The plottype tag value for POLARSCOPE is "polarscope".
 *
 * CARTPLOT	A Cartesian plot. The value string assigned to the object is
 *		parsed as a sequence of (x,y) pairs to be plotted, rather
 *		than just a single point, as in the SCOPE modes (above.)
 *		Thus, each new value string assigned represents an entire plot,
 *		which completely replaces the previous plot.
 *
 *		The plottype tag value for CARTPLOT is "cartplot".
 *
 * POLARPLOT	A polar plot. Like CARTPLOT, except the value string is
 *		taken to be (R, theta) pairs.
 *
 *		The plottype tag value for POLARPLOT is "polarplot".
 *	
 *
 */
enum	plttype	{SCOPE, POLARSCOPE, CARTPLOT, POLARPLOT};	


char*	get_value();
void	sciplot_update();
void	sciplot_sync();
void	parse_vals();


/*
 * Per-instance private parameter struct. Into this structure
 * go parameters that are specific to the new object.
 *
 * The object builder routine creates an instance of this struct
 * for each object of the associated type, fills in the appropriate
 * values, and points to it in the per-object "Objdef" global struct
 * via object->info.
 */
typedef struct sciplotinfo
{
    enum	plttype	type;	/* Plot type. */
    int		lstyle;		/* Line style. */
    int		mkr;		/* Point marker type. */
    float	ymin;		/* Y axis display limit - min */
    float	ymax;		/* Y axis display limit - max */
    char	*xlab;		/* X axis label string. */
    char	*ylab;		/* Y axis label string. */
    char	*legend;	/* Legend string. */
    int		nsamp;		/* # of points (or point-pairs) to display. */ 
    int		list_id;	/* Identifier for data list structure. */
    float	x[NSAMP_MAX];	/* X data. */
    float	y[NSAMP_MAX];	/* Y data. */
} _sciplotinf;

/*
 * Object builder routine for the new object, in this case a sciplot.
 */
void build_sciplot(root, parent)
entry*	root;
Objdef* parent;
{
    int		i, j;			/* Scratch. */
    char	*p, *q;			/* Scratch. */
    int		autoscale;		/* Nonzero if autoscaling. */
    Objdef	*object;		/* SciPlot object descriptor. */
    char	*label;			/* Ptr to object label string. */
    struct	sciplotinfo	*sinf;	/* The per-object parm struct. */
    char	defname[20];		/* Default name constructed here. */
    Widget	sciplot;		/* The Widget Itself. */
    Arg		args[30];		/* Resource args for widget setup. */
    int		narg;			/* Dynaming sizer for args[]. */ 
    int		list_id;		/* SciPlot datalist ID value. */

    static 	int	numsciplot = 1;	/* Counter for naming sciplot objs. */
    
    /*
     * Create the object's global Objdef struct instance.
     * Stuff that goes in here is generic to any and all objects.
     */
    object = new_object(parent);

    /*
     * Create the "custom" info struct for this object. Stuff
     * that goes in here is specific to the type of object being
     * built (in this case, a sciplot).
     */
    sinf = (struct sciplotinfo*) XtMalloc(sizeof(struct sciplotinfo));

    /*
     * Fill in some of the global stuff. 
     */
    object->info = sinf;		/* Our specific stuff. */
    object->updater = sciplot_update;	/* Callback to update screen image. */
    object->sync = sciplot_sync;	/* Callback when our parms are... */
					/* modified by SET. */

    /*
     * Construct a default name for the new object, in case one isn't
     * supplied. The convention is objnameN, where objname is the name
     * of the object, and N is a serialized number, incremented for each
     * new object of this type that is built.
     */
    sprintf(defname, "%s%d", OBJNAME, numsciplot++);
    
    /*----- continued -----*/

    /*
     * Obtain the various parameters (tags) specified in the script for
     * this object, defaulting as appropriate. (get_value() is
     * defined in tree.c)
     *
     * Note that get_value() returns a pointer an XtMalloc-ed static
     * string so in cases where the string itself is not used (e.g.
     * when it is copied or converted into another variable of a
     * different type) the original string should be XtFree-ed.
     * This will neither increase your tax nor reduce your refund.
     */
    object->name = get_value(root, "name", defname);
    p = get_value(root, "marker", "none");
    if (!strcmp(p, "none"))
	sinf->mkr = MKR_NONE;
    else if (!strcmp(p, "dot"))
	sinf->mkr = MKR_DOT;
    else if (!strcmp(p, "circle"))
	sinf->mkr = MKR_CIRC;
    else if (!strcmp(p, "square"))
	sinf->mkr = MKR_SQR;
    else if (!strcmp(p, "triangle"))
	sinf->mkr = MKR_TRI;
    else if (!strcmp(p, "fcircle"))
	sinf->mkr = MKR_FCIRC;
    else if (!strcmp(p, "fsquare"))
	sinf->mkr = MKR_FSQR;
    else if (!strcmp(p, "ftriangle"))
	sinf->mkr = MKR_FTRI;

    p = get_value(root, "linestyle", "solid");
    if (!strcmp(p, "solid"))
	sinf->lstyle = LS_SOLID;
    else if (!strcmp(p, "dotted") || !strcmp(p, "dot"))
	sinf->lstyle = LS_DOT;
    else if (!strcmp(p, "widedotted") || !strcmp(p, "wdot") 
       || !strcmp(p, "widedot"))
	sinf->lstyle = LS_WDOT;
    else if (!strcmp(p, "none"))
	sinf->lstyle = LS_NONE;

    p = get_value(root, "plottype", "scope");
    if (!strcmp(p, "scope") || !strcmp(p, "cartscope"))
	sinf->type = SCOPE;
    else if (!strcmp(p, "cartplot"))
	sinf->type = CARTPLOT;
    else if (!strcmp(p, "polarscope"))
	sinf->type = POLARSCOPE;
    else if (!strcmp(p, "polarplot"))
	sinf->type = POLARPLOT;
    else
    {
	fprintf(stderr,"? unknown plottype '%s' in %s object %s\n",
		 p, OBJNAME, object->name);
	exit(-1);
    }
    XtFree(p);

    label = get_value(root, "label", "");
    sinf->xlab = get_value(root, "xlabel", "");
    sinf->ylab = get_value(root, "ylabel", "");

    sinf->nsamp = ((int) atoi(p=get_value(root, "nsamp",NSAMP_DEFAULT)));
    XtFree(p);

    /*
     * In get_value() calls, '*' 3rd parm means no explicit setting given.
     */
    p = get_value(root, "min", "*");	
    q = get_value(root, "max", "*");

    /*----- continued -----*/

    autoscale = 1;				/* Assume autoscale. */
    sinf->ymin = YMIN_DEFAULT;			/* Assume default min & max. */
    sinf->ymax = YMAX_DEFAULT;
    if (*p != '*')				/* If explicit min specified. */
    {
	sinf->ymin = atof(p);			/* ...then get the value... */
	autoscale = 0;				/* ...and don't autoscale. */
    }

    if (*q != '*')				/* Same as above, for max. */
    {
	sinf->ymax = atof(q);
	autoscale = 0;
    }
    XtFree(p);
    XtFree(q);

    sinf->legend = "";

    /*
     * Some error checking. Complain and exit on nonsense, no need to
     * be anal about it.
     */
    if (!autoscale && (sinf->ymin >= sinf->ymax))
    {
	fprintf(stderr,
	    "? min >= max in %s object %s\n", OBJNAME, object->name);
	exit(-1);
    }

    if (sinf->nsamp < 2 || sinf->nsamp > NSAMP_MAX)
    {
	fprintf(stderr,"? bad nsamp value, %d, in %s object %s\n",
		sinf->nsamp, OBJNAME, object->name);
	exit(-1);
    }

    /*----- continued -----*/

    /*
     * Determine type of sciplot, and extract data from value
     * string accordingly.
     */
    switch(sinf->type)
    {
    case SCOPE:
	/*
	 * Set up X,Y axis values for SCOPE mode. Want X axis to
	 * be "like time", so range of [0, nsamp] will do fine.
	 * Y values are taken to be a sequence of abscissas.
	 *
	 * The 0.01 default value is a workaround for a bug in
	 * SciPlot 1.2, which results in a SING error if 0 is used
	 * as the only point, when autoscaling is on.
	 */
	object->value = get_value(root,"value","0.01");
	parse_vals(object->value, sinf->y, sinf->nsamp, " \t,");
	for (i = 0; i < sinf->nsamp; ++i)
	{
	    sinf->x[i] = i;	/* X axis is just time. */
	}
	break;

    case POLARSCOPE:		/* For these types, expect point pairs. */
    case POLARPLOT:
    case CARTPLOT:
	{
	    float	ftmp[2*NSAMP_MAX];
	    float	*fp = ftmp;
	    /*
	     * The 0.01 default is a workaround for a bug in SciPlot 1.2, 
	     * which results in a SING error if (0,0) is used.
	     */
	    object->value = get_value(root,"value","0.01 0.01");
	    parse_vals(object->value, ftmp, 2*sinf->nsamp, " \t,");
	    for (i = 0; i < sinf->nsamp; ++i)
	    {
		sinf->x[i] = *fp++;
		sinf->y[i] = *fp++;
	    }
	}
	break;


    default:
	fprintf(stderr,"? internal error in %s object %s: bad sinf->type\n",
		OBJNAME, object->name);
	exit(-1);
    }

    /*----- continued -----*/

    /*
     * Set up the SciPlot resources as appropriate for
     * the type of plot display. We're going for simplicity
     * and speed of screen update, so keep it light & fluffy.
     */
    narg = 0;

    XtSetArg(args[narg], XtNyOrigin, True); narg++;
    XtSetArg(args[narg], XtNdegrees, False); narg++;
    XtSetArg(args[narg], XtNshowLegend, False); narg++;
    XtSetArg(args[narg], XtNdrawMajorTics, False); narg++;
    XtSetArg(args[narg], XtNdrawMinor, False); narg++;
    XtSetArg(args[narg], XtNdrawMinorTics, False); narg++;
    switch (sinf->type)
    {
    case POLARSCOPE:
	XtSetArg(args[narg], XtNchartType, XtPOLAR); narg++;
	XtSetArg(args[narg], XtNdrawMajor, False); narg++;
	break; 

    case POLARPLOT:
	XtSetArg(args[narg], XtNchartType, XtPOLAR); narg++;
	XtSetArg(args[narg], XtNdrawMajor, True); narg++;
	break;

    case SCOPE:
	XtSetArg(args[narg], XtNchartType, XtCARTESIAN); narg++;
	XtSetArg(args[narg], XtNdrawMajor, False); narg++;
	break;

    case CARTPLOT:
	XtSetArg(args[narg], XtNchartType, XtCARTESIAN); narg++;
	XtSetArg(args[narg], XtNdrawMajor, True); narg++;
	break;

    default:
	fprintf(stderr,"? internal error in %s object %s: bad sinf->type\n",
		OBJNAME, object->name);
	exit(-1);
    }


    /*
     * The following label/title resources are disabled if
     * the corrsponding string is null.
     */
    XtSetArg(args[narg], XtNplotTitle, label); narg++;
    i = *label ? True : False;
    XtSetArg(args[narg], XtNshowTitle, i); narg++;

    XtSetArg(args[narg], XtNxLabel, sinf->xlab); narg++;
    i = *sinf->xlab ? True : False;
    XtSetArg(args[narg], XtNshowXLabel, i); narg++;

    XtSetArg(args[narg], XtNyLabel, sinf->ylab); narg++;
    i = *sinf->ylab ? True : False;
    XtSetArg(args[narg], XtNshowYLabel, i); narg++;

    /*----- continued -----*/

    /*
     * Grab all the common-to-all-object resources. SET_BEFORE
     * tells common_tags() that we're calling it before creation
     * of the widget, so it doesn't try to set things like visibility
     * resources before the widget exists.
     */
    common_tags(parent->widgetname, root, args, &narg, SET_BEFORE);

    /*
     * Create the widget(s) corresponding to the new object, and
     * copy the Widget descriptor into our global object descriptor.
     */
    sciplot = XtCreateManagedWidget(object->name,
				   sciplotWidgetClass,
				   parent->widgetname,
				   args,
				   narg);
    object->widgetname = sciplot;

    /*
     * Now set up the common resources that can be handled once the widget
     * exists.
     */
    narg = 0;
    common_tags(sciplot, root, args, &narg, SET_AFTER);

    /*
     * Create the SciPlot data list object from the (x,y) data
     */
    sinf->list_id = SciPlotListCreateFromFloat(sciplot,
					 sinf->nsamp, 
					 sinf->x,
					 sinf->y,
					 sinf->legend);

    
    /*
     * If not autoscaling (i.e. either min or max was explicitly specified
     * by user) then turn off autoscaling, use the min and max values
     * specified or defaulted.
     */
    if (!autoscale)
	SciPlotSetYUserScale(sciplot, sinf->ymin, sinf->ymax);
    else
	SciPlotSetYAutoScale(sciplot);

    SciPlotListSetStyle(sciplot, sinf->list_id,
			SciPlotAllocNamedColor(sciplot, "Black"),
				/* Point color. */
			sinf->mkr,
				/* Point style. */
			SciPlotAllocNamedColor(sciplot, "Black"),
				/* Line color. */
			sinf->lstyle);
				/* Line style. */

    SciPlotUpdate(sciplot);
}

/*
 * sciplot_update()
 *
 * Update the screen representation of the SciPlot widget based on
 * the passed "value" string. Called when the sciplot object
 * is ASSIGNed to.
 *
 */
void sciplot_update(object, value)
Objdef	*object;
char	*value;
{
    struct	sciplotinfo	*sinf;

    int		i;
    int		quickupdate = 0;
    float	pairs[2*NSAMP_MAX];
    float	*pp;

    sinf = object->info;

    switch (sinf->type)
    {
    /*
     * In SCOPE mode, value string is just the most recent abcissa.
     * This is appended to the existing dataset, and the oldest one
     * is discarded.
     */
    case SCOPE:
	quickupdate = 1;		/* Use SciPlot fast update mode. */
	for (i = 0; i < sinf->nsamp-1; ++i)
	{
	    sinf->y[i] = sinf->y[i+1];
	}
	sscanf(value, "%f", &sinf->y[i]);
	break;

    /*
     * In POLARSCOPE mode, value string is the most recent (R, theta)
     * pair. This pair is appended to end of existing dataset, and the
     * oldest pair discarded.
     */
    case POLARSCOPE:
	quickupdate = 1;		/* Use SciPlot fast update mode. */ 
	for (i = 0; i < sinf->nsamp-1; ++i)
	{
	    sinf->x[i] = sinf->x[i+1];
	    sinf->y[i] = sinf->y[i+1];
	}
	parse_vals(value, pairs, 2, " \t,");
	sinf->x[i] = pairs[0];
	sinf->y[i] = pairs[1];
	break;

    /*
     * In the PLOT modes, value string is a sequence of (x,y) or
     * (R, theta) pairs which completely replace the existing
     * dataset.
     */
    case CARTPLOT:
    case POLARPLOT:
	parse_vals(value, pairs, 2*sinf->nsamp, " \t,");
	pp = pairs;
	for (i = 0; i < sinf->nsamp; ++i)
	{
	    sinf->x[i] = *pp++;
	    sinf->y[i] = *pp++;
	}
	break;

    default:
	fprintf(stderr,"? internal error in %s object %s: bad sinf->type\n",
		OBJNAME, object->name);
	exit(-1);
    }


    SciPlotListUpdateFromFloat(object->widgetname,
			       sinf->list_id, sinf->nsamp,
			       sinf->x,
			       sinf->y);
    /*
     * Scope modes call the SciPlotQuickUpdate() function which only
     * repaints the data portion, not the axes and labels. So no autoscaling
     * as new data is added.
     */
    if (quickupdate)
	SciPlotQuickUpdate(object->widgetname);
    else
	SciPlotUpdate(object->widgetname);
}

/*
 * sciplot_sync()
 *
 * Called when a sciplot object has a resource changed with SET. 
 * Handle non-common parameters which require special action,
 * such as updating the private data structure or calling
 * functions that update the state of the widget based on the
 * new paramters.
 *
 * Common resources (e.g. height, width) are handled by a call
 * to update_common_tag.
 *
 */
void sciplot_sync(object, tag, value)
Objdef* object;
char* tag;
char* value;
{
    struct	sciplotinfo *sinf = object->info; /* Our private data. */
    int		calluserscale;	/* Set if need to call user scaling. */

    /*
     * Most tags are common to many objects, and are 
     * handled via update_common_tag().
     */
    update_common_tag(object,tag,value);

    /*
     * Handle special tags here. nsamp, min and max.
     */
    if (!strcmp(tag, "nsamp"))
    {
	sinf->nsamp = atoi(value);
	if (sinf->nsamp < 2 || sinf->nsamp > NSAMP_MAX)
	{
	    fprintf(stderr,"? bad nsamp value, %d, ASSIGNed to %s object %s\n",
		    sinf->nsamp, OBJNAME, object->name);
	    exit(-1);
	}
    }

    calluserscale = 0;			/* Assume no min or max specified. */
    if (!strcmp(tag, "min"))
    {
	sinf->ymin =  atof(value);
	calluserscale = 1;
    }
    if (!strcmp(tag, "max"))
    {
	sinf->ymax = atof(value);
	calluserscale = 1;
    }

    if (calluserscale)
	SciPlotSetYUserScale(object->widgetname, sinf->ymin, sinf->ymax);
    /*
     * Always call the screen updater. (If we had been autoscaling, we
     * ain't anymore.)
     */
    SciPlotUpdate(object->widgetname);
}

/* 
 * parse_vals()
 * 
 * Just like parse_valuestring from graph.c. Wanted a separate
 * version for SciPlot use, just in case there some differences
 * arise later.
 *
 */

void parse_vals(valstring, values, nsamp, separator) 
char *valstring;
float *values;
int nsamp;
char *separator;
{
    char *item;
    int isamp,jsamp;

    for( item=strtok(valstring,separator),isamp=0;
         item != (char*)0 && isamp <nsamp ;
         item =strtok( (char*)0, separator ),isamp++){
		sscanf(item,"%f",&values[isamp]);
    }

    /* if valstring does not contain enough samples, reuse the last one */
    if (isamp < nsamp) {
	for (jsamp=isamp; jsamp<nsamp; jsamp++) {
		values[jsamp] = values[isamp-1];
	}
    }
}
